# save all constant variables used by the package
CODE_TYPES = ['tbd']
SPECIAL_TOKEN_DICT = {'tbd':['<tbd>','</tbd>']}

UNKNOWN_TOKEN = '<unk>'

model_max_length = 512

eps = 1e-16

# PRETRAINED_MODEL_URL = 'https://uofi.box.com/shared/static/cu09as2bmotrr9bejsfgumv6yx46mmfw.zip'
PRETRAINED_MODEL_URL = 'https://storage.googleapis.com/pytrial/promptEHR_pretrained.zip'

SYNTHETIC_DATA_URL = 'https://github.com/RyanWangZf/PromptEHR/raw/main/demo_data/synthetic_ehr/data.pkl'


'''
Several utilities for processing tabular datasets.

TODO:

add LabelEncoder, OneHotEncoder, FrequencyEncoder (all encoders for categorical features) that supports

max_categories and min_frequencies, refer to https://scikit-learn.org/stable/modules/preprocessing.html.
'''
from collections import defaultdict
import pickle
from pathlib import Path
import pdb
import os

import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler as sk_standardscaler
from sklearn.preprocessing import MinMaxScaler as sk_minmaxscaler
import rdt
from rdt.hyper_transformer import Config
from rdt.transformers import LabelEncoder, BinaryEncoder, UnixTimestampEncoder
from rdt.transformers.numerical import FloatFormatter

class StandardScaler(FloatFormatter):
    '''Transformer for numerical data.

    This transformer scales all numerical values within the same column using
    `sklearn.preprocessing.StandardScaler`.

    Null values are replaced using a `NullTransformer` from rdt.

    Parameters
    ----------
    missing_value_replacement: object or None
        Indicate what to do with the null values. If an integer or float is given,
        replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
        given, replace them with the corresponding aggregation. If ``None`` is given,
        do not replace them. Defaults to ``None``.

    enforce_min_max_values: bool (default=False)
        Whether or not to clip the data returned by ``reverse_transform`` to the min and
        max values seen during ``fit``. Defaults to ``False``.

    learn_rounding_scheme (bool):
        Whether or not to learn what place to round to based on the data seen during ``fit``.
        If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
        Defaults to ``False``.

    model_missing_values (bool):
        Whether to create a new column to indicate which values were null or not. The column
        will be created only if there are null values. If ``True``, create the new column if
        there are null values. If ``False``, do not create the new column even if there
        are null values. Defaults to ``False``.
    '''
    _dtype = None
    _min_value = None
    _max_value = None

    def __init__(self,
        missing_value_replacement=None,
        enforce_min_max_values=False,
        learn_rounding_scheme=False,
        model_missing_values=None,
        computer_representation='Float'
        ):
        self.missing_value_replacement = missing_value_replacement
        self.enforce_min_max_values=enforce_min_max_values
        self.learn_rounding_scheme=learn_rounding_scheme
        self.model_missing_values = model_missing_values
        self.computer_representation = computer_representation
        self.standard_transformer = sk_standardscaler()

    def _fit(self, data):
        '''
        Fit the transformer to the data.

        Parameters
        ----------
        data: pd.Series
            Data to fit.
        '''
        super()._fit(data)
        data = super()._transform(data)
        self.standard_transformer.fit(data[:,None])

    def _transform(self, data):
        '''
        Transform numerical data.

        Parameters
        ----------
        data: pd.Series
            Data to transform.

        Returns
        -------
            np.ndarray
        '''
        transformed = super()._transform(data)
        transformed = self.standard_transformer.transform(transformed[:,None])
        return transformed.flatten()

    def _reverse_transform(self, data):
        '''
        Convert the transformed data back to the original format.

        Parameters
        ----------
            data: pd.Series or np.ndarray
                Data to be reversely transformed.

        Returns
        -------
            np.ndarray
        '''
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if self.missing_value_replacement is not None:
            data = self.null_transformer.reverse_transform(data)

        if self.enforce_min_max_values:
            data = data.clip(self._min_value, self._max_value)

        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        data = self.standard_transformer.inverse_transform(data[:,None])

        is_integer = np.dtype(self._dtype).kind == 'i'
        if self.learn_rounding_scheme or is_integer:
            data = data.round(self._rounding_digits or 0)

        if pd.isna(data).any() and is_integer:
            return data

        return data.astype(self._dtype)

class MinMaxScaler(StandardScaler):
    '''Transformer for numerical data.

    This transformer scales all numerical values within the same column using
    `sklearn.preprocessing.MinMaxScaler`.

    Null values are replaced using a `NullTransformer` from rdt.

    Parameters
    ----------
    missing_value_replacement: object or None
        Indicate what to do with the null values. If an integer or float is given,
        replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
        given, replace them with the corresponding aggregation. If ``None`` is given,
        do not replace them. Defaults to ``None``.

    enforce_min_max_values: bool (default=False)
        Whether or not to clip the data returned by ``reverse_transform`` to the min and
        max values seen during ``fit``. Defaults to ``False``.

    learn_rounding_scheme (bool):
        Whether or not to learn what place to round to based on the data seen during ``fit``.
        If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
        Defaults to ``False``.

    model_missing_values (bool):
        Whether to create a new column to indicate which values were null or not. The column
        will be created only if there are null values. If ``True``, create the new column if
        there are null values. If ``False``, do not create the new column even if there
        are null values. Defaults to ``False``.
    '''

    _dtype = None
    _min_value = None
    _max_value = None

    def __init__(self,
        missing_value_replacement=None,
        enforce_min_max_values=False,
        learn_rounding_scheme=False,
        model_missing_values=None,
        computer_representation='Float',
        ):
        super().__init__(
            missing_value_replacement=missing_value_replacement,
            enforce_min_max_values=enforce_min_max_values,
            learn_rounding_scheme=learn_rounding_scheme,
            model_missing_values=model_missing_values,
            computer_representation=computer_representation)

        self.standard_transformer = sk_minmaxscaler()


class HyperTransformer(rdt.HyperTransformer):
    '''
    A subclass of `rdt.HyperTransformer` to set special setups.
    '''
    def __init__(self):
        self._default_sdtype_transformers = {
            'numerical': StandardScaler(missing_value_replacement='mean'),
            'categorical': LabelEncoder(),
            'boolean': BinaryEncoder(missing_value_replacement='mode'),
            'datetime': UnixTimestampEncoder(missing_value_replacement='mean'),
        }
        self.field_sdtypes = {}
        self.field_transformers = {}
        self._specified_fields = set()
        self._validate_field_transformers()
        self._valid_output_sdtypes = self._DEFAULT_OUTPUT_SDTYPES
        self._multi_column_fields = self._create_multi_column_fields()
        self._transformers_sequence = []
        self._output_columns = []
        self._input_columns = []
        self._fitted_fields = set()
        self._fitted = False
        self._modified_config = False
        self._transformers_tree = defaultdict(dict)
        self.computer_representation = None

    def detect_initial_config(self, data, verbose=False):
        """Print the configuration of the data.
        This method detects the ``sdtype`` and transformer of each field in the data
        and then prints them as a json object.
        NOTE: This method completely resets the state of the ``HyperTransformer``.
        Args:
            data (pd.DataFrame):
                Data which will have its configuration detected.

            verbose (bool):
                Whether print user message or not.
        """
        # Reset the state of the HyperTransformer
        self.field_sdtypes = {}
        self.field_transformers = {}

        # Set the sdtypes and transformers of all fields to their defaults
        self._learn_config(data)

        if verbose:
            self._user_message('Detecting a new config from the data ... SUCCESS')
            self._user_message('Setting the new config ... SUCCESS')

        config = Config({
            'sdtypes': self.field_sdtypes,
            'transformers': self.field_transformers
        })

        if verbose:
            self._user_message('Config:')
            self._user_message(config)


def read_csv_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                   low_memory=True, encoding=None, index_col=None):
    """Read in csv files with necessary processing
    Parameters
    ----------
    file_loc
    header_lower
    low_memory
    Returns
    -------
    """
    if dtype != None:
        df = pd.read_csv(file_loc, usecols=usecols, dtype=dtype,
                         low_memory=low_memory, encoding=encoding, index_col=index_col)
    else:
        df = pd.read_csv(file_loc, usecols=usecols, low_memory=low_memory,
                         encoding=encoding, index_col=index_col)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_excel_to_df(file_loc, header_lower=True, usecols=None, dtype=None):
    """Read in excel files with necessary processing
    Parameters
    ----------
    file_loc
    header_lower
    low_memory
    Returns
    -------
    """
    if dtype != None:
        df = pd.read_excel(file_loc, usecols=usecols, dtype=dtype)
    else:
        df = pd.read_excel(file_loc, usecols=usecols)
    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_txt_to_df(file_loc, header_lower=True, usecols=None, dtype=None,
                     low_memory=True, encoding=None):
    """Read in excel files with necessary processing

    Parameters
    ----------
    file_loc
    header_lower
    low_memory

    """
    if dtype != None:
        df = pd.read_table(file_loc, usecols=usecols, dtype=dtype,
                           low_memory=low_memory, encoding=encoding)
    else:
        df = pd.read_table(file_loc, usecols=usecols, low_memory=low_memory,
                           encoding=encoding)

    if header_lower:
        df.columns = df.columns.str.lower()
    return df

def read_text_to_list(file_loc, encoding='utf-8'):
    '''
    Read raw text files into a list one line by one line
    '''
    with open(file_loc, 'r', encoding=encoding) as f:
        x = [l.strip() for l in f.readlines()]
    return x


def load_table_config(data_dir, discriminate_bin_feat=False):
    '''
    Load the categorical, numerical, binary feature configuration from the local dataset folder.

    Parameters
    ----------
    data_dir: str
        The target dataset folder.

    discriminate_bin_feat: bool
        Whether discriminate binary and categorical feature or not.

    '''
    num_filepath = os.path.join(data_dir, 'numerical_feature.txt')
    if os.path.exists(num_filepath):
        num_feat_list = read_text_to_list(num_filepath)
    else:
        num_feat_list = []
    
    bin_filepath = os.path.join(data_dir, 'binary_feature.txt')
    if os.path.exists(bin_filepath):
        bin_feat_list = read_text_to_list(bin_filepath)
    else:
        bin_feat_list = []
    bin_feat_list = [c.strip().lower() for c in bin_feat_list]

    # get all column names
    data_filepath = os.path.join(data_dir, 'data_processed.csv')
    df = read_csv_to_df(data_filepath, index_col=0)

    all_feat_list = [c.strip().lower() for c in df.columns.tolist()]
    cat_feat_list = [c for c in all_feat_list if c not in num_feat_list and c not in bin_feat_list and 'target_label' not in c]
    num_feat_list = [c.strip().lower() for c in num_feat_list]

    if not discriminate_bin_feat:
        # merge bin and cat
        cat_feat_list.extend(bin_feat_list)

    # get cat cardinalities
    cat_cardinalities = []
    for c in cat_feat_list:
        cat_cardinalities.append(df[c].nunique())

    return {
        'columns': all_feat_list,
        'num_feat': num_feat_list,
        'bin_feat': bin_feat_list,
        'cat_feat': cat_feat_list,
        'cat_cardinalities': cat_cardinalities,
    }


import abc
import pdb
import os
import json
import math

import torch
from torch import nn
from torch.nn.utils.rnn import pad_sequence

from pytrial.utils.check import check_model_dir
from pytrial.utils.check import make_dir_if_not_exist
from pytrial.tasks.trial_simulation.data import SequencePatient

class SequenceSimulationBase(abc.ABC):
    '''Abstract class for all sequential patient data simulations.
    
    Parameters
    ----------
    experiment_id: str, optional (default = 'test')
        The name of current experiment.
    '''
    training=False
    @abc.abstractmethod
    def __init__(self, experiment_id='test'):
        check_model_dir(experiment_id)
        self.checkout_dir = os.path.join('./experiments_records', experiment_id,
                                         'checkpoints')
        self.result_dir = os.path.join('./experiments_records', experiment_id,
                                       'results')
        make_dir_if_not_exist(self.checkout_dir)
        make_dir_if_not_exist(self.result_dir)
    
    @abc.abstractmethod
    def fit(self, train_data):
        '''Fit the model to the given training data. Need to be implemented by the child class.

        Parameters
        ----------
        train_data: SequencePatient
            The training data.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, number_of_predictions):
        '''Generate synthetic sequence patient data. Need to be implemented by the child class.
        
        Parameters
        ----------
        number_of_predictions: int
            The number of synthetic sequence patient data to generate.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def save_model(self, output_dir):
        '''Save the model to the given directory. Need to be implemented by the child class.

        Parameters
        ----------
        output_dir: str
            The given directory to save the model.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def load_model(self, checkpoint):
        '''Load the model from the given directory. Need to be implemented by the child class.
        
        Parameters
        ----------
        checkpoint: str
            The given directory to load the model.
        '''
        raise NotImplementedError
    
    def train(self, mode=True):
        '''Set the model in training mode. Work similar to `model.train()` in PyTorch.
        
        Parameters
        ----------
        mode: bool, optional (default = True)
            Whether to set the model in training mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.train()
        return self
    
    def eval(self, mode=False):
        '''Set the model in evaluation mode. Work similar to `model.eval()` in PyTorch.

        Parameters
        ----------
        mode: bool, optional (default = False)
            Whether to set the model in evaluation mode.
            ``False`` means the model is in evaluation mode.
            ``True`` means the model is in training mode.
        '''
        self.training = mode
        self.model.eval()
        return self

    def _input_data_check(self, inputs):
        assert isinstance(inputs, SequencePatient), 'Wrong input type.'
    
    def _save_checkpoint(self, state,
                        epoch_id=0,
                        is_best=False,
                        output_dir=None,
                        filename='checkpoint.pth.tar'):
        if output_dir is None:
            output_dir = self.checkout_dir

        if epoch_id < 1:
            filepath = os.path.join(output_dir, 'latest.' + filename)
        elif is_best:
            filepath = os.path.join(output_dir, 'best.' + filename)
        else:
            filepath = os.path.join(self.checkout_dir,
                                    str(epoch_id) + '.' + filename)
        torch.save(state, filepath)

    def _save_config(self, config, output_dir=None):
        if output_dir is None:
            output_dir = self.checkout_dir

        temp_path = os.path.join(output_dir, 'config.json')
        if os.path.exists(temp_path):
            os.remove(temp_path)
        with open(temp_path, 'w', encoding='utf-8') as f:
            f.write(
                json.dumps(config, indent=4)
            )

    def _load_config(self, checkpoint=None):
        '''
        Load model config from the given directory.
        Parameters
        ----------
        checkpoint: str
            The given filepath (e.g., ./checkpoint/config.json)
            to load the model config.
        '''
        if checkpoint is None:
            temp_path = os.path.join(checkpoint, 'config.json')
        else:
            temp_path = checkpoint

        assert os.path.exists(temp_path), 'Cannot find `config.json` under {}'.format(self.checkout_dir)
        with open(temp_path, 'r') as f:
            config = json.load(f)
        return config
    
    def _compute_n_per_sample(self, n_test_sample, n=None, n_per_sample=None):
        if n_per_sample is not None:
            n_total = n_test_sample*n_per_sample
            if n is not None:
                n_total = min(n_total, n)
            return n_total, n_per_sample
        else:
            return n, math.ceil(n_test_sample / n)

    def _get_num_visit(self, data, idx):
        visits = data['v']
        num_visit_list = []
        for k,v in visits.items():
            num_visit_list.append(len(v[idx]))
        
        num_visit_uq = list(set(num_visit_list))
        assert len(num_visit_uq) == 1, f'Find mismatch in the number of visit events {num_visit_list}, please check the input data {visits}.'
        return num_visit_uq[0]

    def _prepare_input(self, data, idx, vdx):
        '''
        Prepare inputs for sequence simulation models.

        Parameters
        ----------
        data: dict[list]
            A single patient records.

        idx: int
            The patient index.

        vdx: int
            The target visit index.

        '''
        visits = data['v']
        feature = data['x']
        if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)

        feature = feature.to(self.device)

        inputs = {
            'v':{},
            'y':{},
            'x':feature[idx], # baseline feature
            }

        for k in visits.keys():
            v = pad_sequence(list(map(torch.tensor,visits[k][idx][:vdx])), 
                    batch_first=True)
            
            y = torch.tensor(visits[k][idx][vdx])

            v = v.to(self.device)
            y = y.to(self.device)

            inputs['v'][k] = v
            inputs['y'][k] = y

        return inputs

class InputEventEmbedding(nn.Module):
    def __init__(self, orders, vocab_size, emb_size, padding_idx) -> None:
        super().__init__()
        # build input embeddings
        emb_dict = {}
        for i, order in enumerate(orders):
            emb_dict[order] = nn.Embedding(vocab_size[i], embedding_dim=emb_size, padding_idx=padding_idx)
        self.embeddings = nn.ModuleDict(emb_dict)
    
    def forward(self, inputs):
        emb_list = []
        for k, v in inputs['v'].items():
            emb = self.embeddings[k](v)
            emb_list.append(emb)
        embs = torch.cat(emb_list, 1)
        return embs

class GAN(nn.Module):
    '''
    Build generative adversarial network module referring to MedGAN.
    '''
    def __init__(self,
        emb_size,
        total_vocab_size,
        gen_dims,
        dis_dims,
        ) -> None:
        super().__init__()

        # generator and discriminator part
        self.gen_module = nn.ModuleDict({
            'fc1': nn.Linear(emb_size*2, gen_dims[0]), # from z+s -> hidden feat
            'fc2': nn.Linear(gen_dims[0], gen_dims[1]),
            'fc3': nn.Linear(gen_dims[1], total_vocab_size),
        })

        self.dis_module = nn.ModuleDict({
            'fc1': nn.Linear(total_vocab_size, dis_dims[0]), # x,\bar{x} -> {0,1}
            'fc2': nn.Linear(dis_dims[0], 1),
        })

    def forward(self, s):
        '''forward makes generation taking the state embeddings as inputs.
        '''
        z_random = torch.randn(s.size()).to(s.device)
        return self.infer_generator(s, z_random)
        
    def infer_generator(self, s, z):
        '''do generator inference, x
        args:
            s: state vector [bs, emb_size]
            z: visit or random [bs, emb_size]
        '''
        inputs = torch.cat([z,s], -1) # bs, emb_size*2
        tempVec = inputs
        h = self.gen_module['fc1'](tempVec) # bs, 256
        h = nn.ReLU()(h)

        tempVec = h + tempVec

        h = self.gen_module['fc2'](tempVec)
        h = nn.ReLU()(h)
        tempVec = h + tempVec # bs, 256

        h = self.gen_module['fc3'](tempVec)
        outputs = nn.Sigmoid()(h)
        return outputs

    def infer_discriminator(self, x):
        '''do discriminator inference
        args:
            x: the decoded fake records or the true records x
        '''
        h = self.dis_module['fc1'](x)
        h = nn.ReLU()(h)
        h = self.dis_module['fc2'](h)
        pred = h.sigmoid().squeeze(1)
        return pred


class RNN(nn.Module):
    RNN_TYPE = {
        'rnn':nn.RNN,
        'lstm':nn.LSTM,
        'gru':nn.GRU,
    }
    def __init__(self, 
        rnn_type,
        emb_size,
        num_layer,
        bidirectional,
        ) -> None:
        super().__init__()
        self.model = self.RNN_TYPE[rnn_type](
            input_size=emb_size,
            hidden_size=emb_size,
            num_layers=num_layer,
            bidirectional=bidirectional,
            batch_first=True,
        )
        self.bidirectional = bidirectional
    
    def forward(self, x):
        outputs = self.model(x)[0]
        return outputs


from collections import defaultdict
import pdb
import os
from typing import Any, Callable, Dict, List, NewType, Optional, Tuple, Union

import pandas as pd
from torch.utils.data import Dataset, DataLoader

from ..utils.trial_utils import ClinicalTrials
from ..utils.tabular_utils import read_csv_to_df
from ..data.vocab_data import Vocab

class TrialDatasetBase(Dataset):
    '''
    The basic trial datasets loader.

    Parameters
    ----------
    data: pd.DataFrame
        Contain the trial document in tabular format.
    '''
    inc_ec_embedding = None # inclusion criteria embedding
    inc_vocab = None # inclusion criteria vocab
    exc_ec_embedding = None # exclusion criteria embedding
    exc_vocab = None # exclusion criteria vocab

    def __init__(self, data):
        self.df = data
        self._process_ec()
        self._collect_cleaned_sentence_set()
    
    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        return self.df.iloc[index:index+1]
    
    def get_ec_sentence_embedding(self):
        '''
        Process the eligibility criteria of each trial,
        get the criterion-level emebddings stored in dict.
        '''
        self._get_ec_emb()

    def _process_ec(self):
        res = self.df['criteria'].apply(lambda x: self._split_protocol(x))
        self.df['inclusion_criteria'] = res.apply(lambda x: x[0])
        self.df['exclusion_criteria'] = res.apply(lambda x: x[1])

    def _get_ec_emb(self):
        # create EC embedding with indexed ECs
        from pytrial.model_utils.bert import BERT
        bert_model = BERT(device='cuda:0')
        self.inc_ec_embedding = bert_model.encode(self.inc_vocab.words, batch_size=64)
        self.exc_ec_embedding = bert_model.encode(self.exc_vocab.words, batch_size=64)
        self.inc_ec_embedding = self.inc_ec_embedding.cpu()
        self.exc_ec_embedding = self.exc_ec_embedding.cpu()
        

    def _collect_cleaned_sentence_set(self):
        # create a vocab for ec sentences
        self.inc_vocab = Vocab()
        self.exc_vocab = Vocab()
        self.inc_vocab.add_sentence(['[PAD]']) # 0 belongs to the pad token
        self.exc_vocab.add_sentence(['[PAD]']) # 0 belongs to the pad token

        inc_index_set, exc_index_set = [], []
        for idx, row in self.df.iterrows():
            row_inc_set, row_exc_set = [], []
            inc = row['inclusion_criteria']
            exc = row['exclusion_criteria']
            for sent in inc:
                self.inc_vocab.add_sentence(sent)
                row_inc_set.append(self.inc_vocab.word2idx[sent])
            for sent in exc:
                self.exc_vocab.add_sentence(sent)
                row_exc_set.append(self.exc_vocab.word2idx[sent])
            inc_index_set.append(list(set(row_inc_set)))
            exc_index_set.append(list(set(row_exc_set)))
        self.df['inclusion_criteria_index'] = inc_index_set
        self.df['exclusion_criteria_index'] = exc_index_set

    def _clean_protocol(self, protocol):
        protocol = protocol.lower()
        protocol_split = protocol.split('\n')
        filter_out_empty_fn = lambda x: len(x.strip())>0
        strip_fn = lambda x: x.strip()
        protocol_split = list(filter(filter_out_empty_fn, protocol_split))	
        protocol_split = list(map(strip_fn, protocol_split))
        return protocol_split

    def _split_protocol(self, protocol):
        protocol_split = self._clean_protocol(protocol)
        inclusion_idx, exclusion_idx = len(protocol_split), len(protocol_split)	
        for idx, sentence in enumerate(protocol_split):
            if "inclusion" in sentence:
                inclusion_idx = idx
                break
        for idx, sentence in enumerate(protocol_split):
            if "exclusion" in sentence:
                exclusion_idx = idx 
                break 		
        if inclusion_idx + 1 < exclusion_idx + 1 < len(protocol_split):
            inclusion_criteria = protocol_split[inclusion_idx:exclusion_idx]
            exclusion_criteria = protocol_split[exclusion_idx:]
            if not (len(inclusion_criteria) > 0 and len(exclusion_criteria) > 0):
                print(len(inclusion_criteria), len(exclusion_criteria), len(protocol_split))
                exit()
            return inclusion_criteria, exclusion_criteria ## list, list 
        else:
            return protocol_split, []


class TrialDataset(Dataset):
    '''
    Basic trial datasets loader.

    Parameters
    ----------
    input_dir: str
        The path to the trial dataset in tabular form (.csv).
        If a directory is given, the code will automatically pick the only '.csv' file under this dir.
    '''
    def __init__(self, input_dir=None) -> None:
        if os.path.isfile(input_dir):
            self.df = read_csv_to_df(input_dir, index_col=0)

        if os.path.isdir(input_dir):
            csv_names = [name for name in os.listdir(input_dir) if name.endswith('.csv')]
            if len(csv_names) > 1:
                raise Exception(f'`input_dir` {input_dir} is given where more than one csv files are found under this path.')
            if len(csv_names) == 0:
                raise Exception(f'`input_dir` {input_dir} is given where no csv file is found under this path.')
            self.df = read_csv_to_df(os.path.join(input_dir, csv_names[0]), index_col=0)

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        return self.df.iloc[index:index+1]

class TrialDataCollator:
    '''The basic trial data collator.
    Subclass it and override the `__init__` & `__call__` function if need operations inside this step.

    Returns
    -------
    batch_df: pd.DataFrame
        A dataframe contains multiple fields for each trial.
    '''
    def __init__(self) -> None:
        # subclass to add tokenizer
        # subclass to add feature preprocessor
        pass

    def __call__(self, examples):
        batch_df = pd.concat(examples, 0)
        batch_df.fillna('none',inplace=True)
        return batch_df


class TrialOutcomeDatasetBase(Dataset):
    '''
    Basic trial outcome datasets loader.

    Parameters
    ----------
    data: pd.DataFrame
        Contain the trial document in tabular format.
    '''
    columns = ['nctid', 'label', 'smiless',  'icdcodes', 'criteria']
    def __init__(self, data, columns=None) -> None:
        self.data = data
        if columns is not None:
            self.columns = columns
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        row = self.data.iloc[index]
        return row[self.columns[0]], row[self.columns[1]], row[self.columns[2]], row[self.columns[3]], row[self.columns[4]]

def test():
    trialdata = TrialDataset('./datasets/AACT-ClinicalTrial/')
    trial_collate_fn = TrialDataCollator()
    trialoader = DataLoader(trialdata, batch_size=10, shuffle=False, collate_fn=trial_collate_fn)
    batch = next(iter(trialoader))
    print(batch)

if __name__ == '__main__':
    test()



'''
Evaluation methods for the simulated sequence of patient records.
'''
import pdb
from copy import deepcopy

from torch.utils.data import DataLoader
from sklearn.metrics import roc_auc_score

from pytrial.tasks.indiv_outcome.sequence import RNN
from pytrial.data.patient_data import SeqPatientCollator
from pytrial.data.utils import concat_dataset, split_dataset
from pytrial.tasks.indiv_outcome.trainer import IndivSeqTrainer
from ..data import SequencePatient

__all__ = ['RNNPrivacyDetection', 'RNNUtilityDetection']


class SequenceMetric:
    '''
    Base class for evaluation of the simulated sequence of patient records.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        raise NotImplementedError

    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''
        raise NotImplementedError
    
    @classmethod
    def compute(self, real_data, syn_data):
        raise NotImplementedError

class SequencePrivacyMetric(SequenceMetric):
    '''
    Base class for Machine Learning detection based metrics for sequential synthetic patient records.

    These metrics build a Machine Learning Classifier that learns to tell the synthetic
    data apart from the real data.

    The output is the binary classification AUC.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        Build a dataset for training a ML model.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        Returns
        -------
        train_set: torch.utils.data.Dataset
            A dataset for training a ML model.
        
        test_set: torch.utils.data.Dataset
            A dataset for testing a ML model.
        '''
        real_syn_data = concat_dataset(real_data, syn_data)
        train_data, test_data = split_dataset(real_syn_data, split=0.8)
        return train_data, test_data
    
    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''      
        raise NotImplementedError


class RNNPrivacyDetection(SequencePrivacyMetric):
    '''
    Try to discriminate real and synthetic patient records using an RNN based neural network.
    The output score is AUC evaluated on the prediction.
    So the biggest score will be AUC=1.0, which means the model can correctly discriminate real and synthetic data.
    The lowest score will be AUC=0.5, which means the model cannot discriminate real and sytnethic data.

    There are multiple interpretations of the score. 
    A low score (low AUC) can indicates high synthetic data quality as well as low privacy. 
    A high score (high AUC) can indicate low synthetic data quality as well as high privacy. 
    '''

    @classmethod
    def compute(self, real_data, syn_data, device='cpu'):
        '''
        Compute the detection score.

        Parameters
        ----------
        real_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.data.patient_sequence.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        device: str
            The device to run the model on. Default is 'cpu'.

        Returns
        -------
        float: The detection score.
        '''
        assert isinstance(real_data, SequencePatient), '`real_data` must be a `SequencePatient` object.'
        assert isinstance(syn_data, SequencePatient), '`syn_data` must be a `SequencePatient` object.'

        # assign labels to real and syn data
        real_data = deepcopy(real_data)
        syn_data = deepcopy(syn_data)
        real_data.label = [1] * len(real_data) # 1 is real data label
        syn_data.label = [0] * len(syn_data) # 0 is syn data label

        # build a dataset with labeled real and syndata
        train_data, test_data = self._build_dataset(real_data, syn_data)

        # train a RNN model
        model = self._build_model(train_data, test_data, device=device)

        # evaluate the model
        pred = model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        return auc

    @staticmethod
    def _build_model(train_data, test_data, device='cpu'):
        '''
        Build a RNN model for binary classification.

        Parameters
        ----------
        dataloader: torch.utils.data.DataLoader
            A dataloader for training a ML model.

        Returns
        -------
        A ML model for binary classification of synthetic and real patient records.
        '''
        model = RNN(
            vocab_size=[len(train_data.metadata['voc'][o]) for o in train_data.metadata['visit']['order']],
            orders=train_data.metadata['visit']['order'],
            mode='binary',
            bidirectional=False,
            epochs=10,
            batch_size=64,
            device=device,
            )
        model.fit(train_data=train_data, valid_data=test_data)
        return model


class SequenceUtilityMetric(SequenceMetric):
    '''
    Base class for Machine Learning based models to learn from the synthetic patient records.
    And predict the outcome of the real patient records.
    Compared with the model learning from the real patient records.
    The output is the AUC of the prediction from the real-data-model and the synthetic-data-model.
    '''
    @staticmethod
    def _build_dataset(real_data, syn_data):
        '''
        Build a dataset for training a ML model.

        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        Returns
        -------
        train_set: torch.utils.data.Dataset
            A dataset for training a ML model.
        
        test_set: torch.utils.data.Dataset
            A dataset for testing a ML model.
        '''
        train_real_data, test_real_data = split_dataset(real_data, split=0.7)
        train_syn_data, test_syn_data = split_dataset(syn_data, split=0.7)
        return [(train_real_data, test_real_data), (train_syn_data, test_syn_data)]
    
    @staticmethod
    def _build_model(train_data, test_data):
        '''
        Build a detection model.
        '''
        raise NotImplementedError


class RNNUtilityDetection(SequenceUtilityMetric):
    @staticmethod
    def _build_model(train_data, test_data, device='cpu'):
        '''
        Build a detection model.
        '''
        model = RNN(
            vocab_size=[len(train_data.metadata['voc'][o]) for o in train_data.metadata['visit']['order']],
            orders=train_data.metadata['visit']['order'],
            mode='binary',
            bidirectional=False,
            epochs=50,
            batch_size=128,
            weight_decay=0,
            rnn_type='lstm',
            n_rnn_layer=1,
            learning_rate=5e-4,
            device=device,
            )
        model.fit(train_data=train_data, valid_data=test_data)
        return model

    @classmethod
    def compute(self, real_data, syn_data, device='cpu'):
        '''
        Compute the detection score. Report the performance of the model trained on real data (real-data-model) 
        and the model trained on synthetic data (syn-data-model).
        Performance is evaluated on the test set of the real data.

        If the syn-data-model performs equally well or better than the real-data-model, then the synthetic data has high utility.
        
        Parameters
        ----------
        real_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence object that contains real patient records.

        syn_data: pytrial.tasks.trial_simulation.data.PatientSequence
            A PatientSequence data structure containing the synthetic patient data.

        device: str
            The device to run the model. Default is 'cpu'.

        Returns
        -------
        float: The detection score of real-data-model and synthetic-data-model.
        '''
        assert isinstance(real_data, SequencePatient), '`real_data` must be a `SequencePatient` object.'
        assert isinstance(syn_data, SequencePatient), '`syn_data` must be a `SequencePatient` object.'

        # build a dataset with labeled real and syndata
        real_data_traintest, syn_data_traintest = self._build_dataset(real_data, syn_data)

        # train a RNN model
        real_model = self._build_model(real_data_traintest[0], real_data_traintest[1])
        syn_model = self._build_model(syn_data_traintest[0], syn_data_traintest[1], device=device)

        # evaluate the model
        outputs = {'real-data-model-auc': None, 'syn-data-model-auc': None}
        test_data = real_data_traintest[1] 

        pred = real_model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        outputs['real-data-model-auc'] = auc

        pred = syn_model.predict(test_data)
        label = test_data.get_label()
        auc = roc_auc_score(label, pred)
        outputs['syn-data-model-auc'] = auc

        return outputs